/* 
   Bordeaux : 1995
   */

#include "ext.h"

#include "adjust.h"
#include "fa_utils.h"


/* generic function to build the NFAs corresponding at the empty language 
   and the empty word language 
   bfinal = TRUE if the state must be final (epsilon language) */
static nfa generic_eps_and_empty_nfa(posint size_al, boole bfinal);

/* generic function to build the NFAs corresponding at the one_letter language 
   and the alphabet language :
   in the first case, letterno corresponds to the letter
   else letterno must be initialized to size_al + 1 */
static nfa generic_one_letter_and_alphabet_nfa(posint size_al, posint letterno);

/* generic function to build the DFAs corresponding at the one_letter language 
   and the alphabet language :
   in the first case, letterno corresponds to the letter
   else letterno must be initialized to size_al + 1 */
static dfa generic_one_letter_and_alphabet_dfa(posint size_al, posint letterno);

/* generic function for test if an automaton (represented by its array of final
   (and initial if NFA) states (final) and its number of states (qno))corresponds 
   to the empty language */
static boole generic_test_empty_fa(posint qno, mrkfin final);

/* macro-function used to init the common fields of the DFA and NFA structure
   used in fa_utils_init_fields_dfa and fa_utils_init_fields_nfa */
#define INIT_COMMON_FIELDS(_FA_, _F1_, _F2_, _F3_, _F4_) \
   { \
      (_FA_)->qno     = (_F1_); \
      (_FA_)->sno     = (_F2_); \
      (_FA_)->delta   = (_F3_); \
      (_FA_)->minimal = (_F4_); \
   } \



/* ---------------------------- Global functions ------------------------------- */

void fa_utils_init_fields_dfa(dfa X_dfa, posint qno, posint init, posint sno, 
			      mrkfin final, ddelta delta, boole minimal)
{
  if ( X_dfa != NULL )
    {
      INIT_COMMON_FIELDS(X_dfa, qno, sno, delta, minimal);
      X_dfa->final = final;
      X_dfa->init  = init;
    }  
  else
    perror("fa_utils_init_fields_dfa");
}


void fa_utils_init_fields_nfa(nfa X_nfa, posint qno, posint sno, mrkfin final,
			      ndelta delta, boole minimal, boole is_eps)
{
  if ( X_nfa != NULL )
    {
      INIT_COMMON_FIELDS(X_nfa, qno, sno, delta, minimal);
      X_nfa->infin  = final;
      X_nfa->is_eps = is_eps;
    }
  else
    perror("fa_utils_init_fields_nfa");
}


void fa_utils_free_temporary_fa(boole bnfa, int nb_arg, ...)
{
  va_list fa_list;
  int i;

  va_start(fa_list, nb_arg);
  
  if( bnfa == FALSE )
    for ( i = 1 ; i <= nb_arg ; i++ )
      {
	dfa tmp_dfa;
	
	tmp_dfa = va_arg(fa_list, dfa);
	
	freedfa(tmp_dfa);
      }
  else
    for ( i = 1 ; i <= nb_arg ; i++ )
      {
	nfa tmp_nfa;
	
	tmp_nfa = va_arg(fa_list, nfa);
	
	freenfa(tmp_nfa);
      }
  
  va_end(fa_list);
}


dfa fa_utils_duplicate_dfa(dfa X_dfa)
{
  dfa    result_dfa;
  mrkfin X_final;
  ddelta X_delta;
  posint result_qno;
  posint result_sno;
  int    i, j;

  result_dfa = newdfa();

  result_qno = X_dfa->qno;
  result_sno = X_dfa->sno;

  fa_utils_init_fields_dfa(result_dfa, result_qno, X_dfa->init, result_sno, 
			   newfinal(result_qno), newddelta(result_sno, result_qno), 
			   X_dfa->minimal);

  X_final = X_dfa->final;
  X_delta = X_dfa->delta;
      
  for ( i = 0 ; i <= result_qno ; i++ )
    result_dfa->final[i] = X_final[i];

  for( i = 1 ; i <= result_sno ; i++ )
    for ( j = 0 ; j <= result_qno ; j++ )
      result_dfa->delta[i][j] = X_delta[i][j];

  return result_dfa;
}



dfa fa_utils_prefix_i_dfa(dfa X_dfa, posint newfinal)
/* The only final state of the result DFA is newfinal : the 
   minimalisation makes that all of the non reachable states 
   are eliminated of the result DFA */
{
  dfa result_dfa;

  result_dfa = fa_utils_duplicate_dfa(X_dfa);

  /* if the automaton has only 1 state : no changes */
  if (X_dfa->qno != 0)  
    {
      posint qno;
      int    i;

      qno  = result_dfa->qno;

      for ( i = 0 ; i <= qno ; i++ )
	setfinalF(result_dfa->final[i]);       
      setfinalT(result_dfa->final[newfinal]);  /* newfinal is marked as final */
      
      result_dfa = dfamdfa(result_dfa, TRUE);
    }

  return result_dfa;
}




dfa fa_utils_suffix_i_dfa(dfa X_dfa, posint newinit)
/* The initial state of the result DFA is newinit : the 
   minimalisation makes that all of the non reachable states 
   are eliminated of the result DFA */
{
  dfa result_dfa;

  result_dfa = fa_utils_duplicate_dfa(X_dfa);

  /* if the automaton has only 1 state : no changes */
  if (X_dfa->qno != 0)
    {
      result_dfa->init = newinit;     /* newinit is the new initial state */      
      result_dfa = dfamdfa(result_dfa, TRUE);
    }

  return result_dfa;
}



dfa fa_utils_empty_dfa(posint size_al)
/* Thew empty DFA is a DFA with one state that is not final */
{
  dfa    result_dfa;
  ddelta delta;
  mrkfin final;
  int    i;

  result_dfa = newdfa();
  delta      = newddelta(size_al, 0); /* 0 : one state */
  final      = newfinal(0);

  setfinalF(final[0]);

  for ( i = 1 ; i <= size_al ; i++ )
    delta[i][0] = 0;

  fa_utils_init_fields_dfa(result_dfa, 0, 0, size_al, final, delta, TRUE);

  return result_dfa;
}


nfa fa_utils_empty_nfa(posint size_al)
/* The empty language NFA is a NFA with one state (not final) and no transition
   exiting from this state */
{
  return generic_eps_and_empty_nfa(size_al, FALSE);
}


dfa fa_utils_eps_dfa(posint size_al)
/* The epsilon DFA is a DFA with two states of this form :

   ->(*0*)----------------->(well)--------- 
            (a_1 ... a_k)     /|\         |(a_1 ... a_k)
                               |          |
                               ------------
   where k = size_al, (*i*) marks a final state, */
{
  dfa    result_dfa;
  ddelta delta;
  mrkfin final;
  int    i;

  result_dfa = newdfa();
  delta      = newddelta(size_al, 1); /* 1 : two states */
  final      = newfinal(1);

  setfinalT(final[0]);
  setfinalF(final[1]);

  for ( i = 1 ; i <= size_al ; i++ )
    {
      delta[i][0] = 1;
      delta[i][1] = 1;
    }

  fa_utils_init_fields_dfa(result_dfa, 1, 0, size_al, final, delta, TRUE);

  return result_dfa;
}



nfa fa_utils_eps_nfa(posint size_al)
/* The empty language NFA is a NFA with one state (final) and no transition
   exiting from this state */
{
  return generic_eps_and_empty_nfa(size_al, TRUE);
}


dfa fa_utils_one_letter_dfa(posint size_al, posint letterno)
/* The one_letter DFA is a DFA with three states of this form :

   ->( 0 )--------------->(*1*)------------------->(well)---------- 
       |   (a_letterno)          (a_1 ... a_k)     /|\/|\         |(a_1 ... a_k)
       |                                            |  |          |
       ----------------------------------------------  ------------
                     (a_1 ... a_k)

   where k = size_al, (*i*) mark a final state */
{
  return generic_one_letter_and_alphabet_dfa(size_al, letterno);
}


nfa fa_utils_one_letter_nfa(posint size_al, posint letterno)
/* The one_letter NFA is a NFA of this form :

   ->( 0 )--------------->(*1*)
           (a_letterno)        

   (*i*) mark a final state */
{
  return generic_one_letter_and_alphabet_nfa(size_al, letterno);
}



dfa fa_utils_alphabet_dfa(posint size_al)
/* The alphabet DFA is a DFA with three states of this form :

   ->( 0 )--------------->(*1*)----------------->(well)-------- 
           (a_1 ... a_k)         (a_1 ... a_k)      /|\       | (a_1 ... a_k)
                                                     |        |
                                                     ----------
   where k = size_al, (*i*) mark a final state */
{
  return generic_one_letter_and_alphabet_dfa(size_al, size_al + 1);
}

nfa fa_utils_alphabet_nfa(posint size_al)
/* The alphabet NFA is a NFA with two states of this form :

   ->( 0 )--------------->(*1*)
           (a_1 ... a_k)       

   where k = size_al, (*i*) mark a final state */
{
  return generic_one_letter_and_alphabet_nfa(size_al, size_al + 1);
}



boole fa_utils_test_empty_dfa(dfa X_dfa)
{
  return generic_test_empty_fa(X_dfa->qno, X_dfa->final);
}


boole fa_utils_test_empty_nfa(nfa X_nfa)
{
  return generic_test_empty_fa(X_nfa->qno, X_nfa->infin);
}

boole fa_utils_test_eps_dfa(dfa X_dfa)
/* A minimal automaton represents the empty word if it has
   - a first state that is initial and final 
   - no transition from the initial state to itself
   - a well state
   */
{
  boole  result;

  result  = FALSE;

  if ( X_dfa->qno == 1 )     /* X_dfa has only 2 states : 0 and 1 */
    {
      posint init;  /* initial state */
      posint other; /* other state */
      posint sno;
      
      init = X_dfa->init;
      sno  = X_dfa->sno;

      if ( init == 0 )
	other = 1;
      else
	other = 0;

      if ( isfinal(X_dfa->final[init]) != TRUE 
	  || isfinal(X_dfa->final[other]) == TRUE )
	result = FALSE;
      else /* we search if there is a "loop" on the initial state */
	{
	  int i;
	  
	  result = TRUE;
	  for( i = 1 ; i <= sno ; i++ )
	    if( X_dfa->delta[i][init] == init ) 
	      /* there is a transition from the initial state to itself  */
	      {
		result = FALSE;
		break;
	      }
	}
    }

  return result;
}


boole fa_utils_test_if_eps_in_dfa(dfa X_dfa)
/* eps is recognized by X_dfa if the initial state of X_dfa is final */
{
  return( isfinal(X_dfa->final[X_dfa->init]) == TRUE );
}

boole fa_utils_test_if_eps_in_nfa(nfa X_nfa)
/* eps is recognized by X_nfa if an initial state of X_nfa is final 
   X_nfa has no epsilon transition */
{
  posint i;
  posint qno;
  mrkfin infin;

  qno   = X_nfa->qno;
  infin = X_nfa->infin;

  for( i = 0 ; i <= qno ; i++ )
    if( isfinal(infin[i]) == TRUE && isinit(infin[i]) == TRUE )
      return TRUE;

  return FALSE;
}


boole fa_utils_test_sub_monoid(dfa X_dfa, nfa X_nfa)
/* the language X (with X_dfa as minimal DFA and X_nfa as NFA)
   is a sub-monoid iff X = (X)* */
{
  dfa   star_X_dfa;
  nfa   star_X_nfa;
  boole result;
  
  star_X_nfa = adjust_star_nfa(X_nfa);
  star_X_dfa = dfamdfa(nfa2dfa(star_X_nfa), TRUE);
  result = equiv(X_dfa, star_X_dfa);

  freenfa(star_X_nfa);
  freedfa(star_X_dfa);
  return result;
}


posint fa_utils_well(dfa X_dfa)
/* X_dfa must be minimal, else it could have several well states.
   A state is the well if all the transitions that leave this 
   state comes back to this same state (direct loops on this 
   state) */
{
  int    qno;
  int    sno;
  mrkfin final;
  ddelta delta;
  int    i, j;

  qno   = X_dfa->qno;
  sno   = X_dfa->sno;
  final = X_dfa->final;
  delta = X_dfa->delta;

  for ( i = 0 ; i <= qno ; i++ )
    {
      if ( !isfinal(final[i]) ) /* a final state can't be the well */
	{
	  for ( j = 1 ; (j <= sno) && (delta[j][i] == i) ; j++ ) ;
	  if ( j == sno + 1 )/* no transitions has gone to an other state than i */
	    return i;
	}
    }

  return (qno + 1); /* X_dfa has no well state (A*) */
}


/* ---------------------------- Static functions ------------------------------- */


static nfa generic_eps_and_empty_nfa(posint size_al, boole bfinal)
/* the empty language and empty word language NFAs have 1 state */
{
  nfa    result_nfa;
  ndelta delta;
  mrkfin infin;
  int    i;

  result_nfa = newnfa();
  delta      = newndelta(size_al, 0); /* 0 : one state */
  infin      = newfinal(0);

  setinit(infin[0]);
  setfinal(infin[0], bfinal);
  
  fa_utils_init_fields_nfa(result_nfa, 0, size_al, infin, delta, TRUE, FALSE);

  return result_nfa;
}


static nfa generic_one_letter_and_alphabet_nfa(posint size_al, posint letterno)
/* the one_letter and alphabet NFAs have 2 states */
{
  nfa    result_nfa;
  ndelta delta;
  mrkfin infin;

  result_nfa = newnfa();
  delta      = newndelta(size_al, 1); /* 1 : two states */
  infin      = newfinal(1);

  setfinalF(infin[0]);
  setinit(infin[0]);
  setfinalT(infin[1]);

  if( letterno <= size_al)
    connect(delta, letterno, 0, 1); /* transition from 0 to 1 by letter letterno */
  else
    {
      int i;

      for( i = 1 ; i <= size_al ; i++ )
	connect(delta, i, 0, 1);
    } 
  
  fa_utils_init_fields_nfa(result_nfa, 1, size_al, infin, delta, TRUE, FALSE);

  return result_nfa;
}


static dfa generic_one_letter_and_alphabet_dfa(posint size_al, posint letterno)
{
  dfa    result_dfa;
  ddelta delta;
  mrkfin final;
  int    i;

  result_dfa = newdfa();
  delta      = newddelta(size_al, 2);  /* 2 : three states */
  final      = newfinal(2);

  setfinalF(final[0]);
  setfinalT(final[1]);
  setfinalF(final[2]);

  for (i = 1 ; i <= size_al ; i++ )
    {
      if( letterno == (size_al + 1) )
	delta[i][0] = 1;
      else
	delta[i][0] = 2;

      delta[i][1] = 2;
      delta[i][2] = 2;
    }

  if( letterno != (size_al + 1) )
    delta[letterno][0] = 1;

  fa_utils_init_fields_dfa(result_dfa, 2, 0, size_al, final, delta, TRUE);

  return result_dfa;
}


static boole generic_test_empty_fa(posint qno, mrkfin final)
/* an automaton corresponds to the empty language if none state is final */
{
  posint i;

  for( i = 0 ; i<= qno ; i++ )
    if( isfinal(final[i]) == TRUE )
      return FALSE;

  return TRUE;
}
